/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef STDC_HEADERS
# include <stdio.h>
#endif
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include <mysql.h>
#include <mysqld_error.h>

#include "sb_options.h"
#include "db_driver.h"

/* Check MySQL version for prepared statements availability */
#if MYSQL_VERSION_ID >= 40103
# define HAVE_PS
#endif

/* MySQL table types */

typedef enum
{
  TABLE_TYPE_MYISAM,
  TABLE_TYPE_INNODB,
  TABLE_TYPE_BDB,
  TABLE_TYPE_HEAP,
  TABLE_TYPE_NDB,
  TABLE_TYPE_FEDERATED
} mysql_drv_table_t;

/* MySQL driver arguments */

static sb_arg_t mysql_drv_args[] =
{
  {"mysql-host", "MySQL server host", SB_ARG_TYPE_STRING, "localhost"},
  {"mysql-port", "MySQL server port", SB_ARG_TYPE_INT, "3306"},
  {"mysql-socket", "MySQL socket", SB_ARG_TYPE_STRING, NULL},
  {"mysql-user", "MySQL user", SB_ARG_TYPE_STRING, "sbtest"},
  {"mysql-password", "MySQL password", SB_ARG_TYPE_STRING, ""},
  {"mysql-db", "MySQL database name", SB_ARG_TYPE_STRING, "sbtest"},
  {"mysql-table-type", "type of test table {myisam,innodb,bdb,heap,ndbcluster,federated}",
   SB_ARG_TYPE_STRING, "innodb"},
  {"myisam-max-rows", "max-rows parameter for MyISAM tables", SB_ARG_TYPE_INT, "1000000"},
  
  {NULL, NULL, SB_ARG_TYPE_NULL, NULL}
};

typedef struct
{
  char               *host;
  unsigned int       port;
  char               *socket;
  char               *user;
  char               *password;
  char               *db;
  mysql_drv_table_t  table_type;
  unsigned int       myisam_max_rows;
} mysql_drv_args_t;

#ifdef HAVE_PS
/* Structure used for DB-to-MySQL bind types map */

typedef struct
{
  db_bind_type_t   db_type;
  int              my_type;
} db_mysql_bind_map_t;

/* DB-to-MySQL bind types map */
db_mysql_bind_map_t db_mysql_bind_map[] =
{
  {DB_TYPE_TINYINT,   MYSQL_TYPE_TINY},
  {DB_TYPE_SMALLINT,  MYSQL_TYPE_SHORT},
  {DB_TYPE_INT,       MYSQL_TYPE_LONG},
  {DB_TYPE_BIGINT,    MYSQL_TYPE_LONGLONG},
  {DB_TYPE_FLOAT,     MYSQL_TYPE_FLOAT},
  {DB_TYPE_DOUBLE,    MYSQL_TYPE_DOUBLE},
  {DB_TYPE_DATETIME,  MYSQL_TYPE_DATETIME},
  {DB_TYPE_TIMESTAMP, MYSQL_TYPE_TIMESTAMP},
  {DB_TYPE_CHAR,      MYSQL_TYPE_STRING},
  {DB_TYPE_VARCHAR,   MYSQL_TYPE_VAR_STRING},
  {DB_TYPE_NONE,      0}
};

#endif /* HAVE_PS */

/* MySQL driver capabilities */

static drv_caps_t mysql_drv_caps =
{
  .multi_rows_insert = 1,
  .transactions = 0,
  .prepared_statements = 0,
  .auto_increment = 1,
  .serial = 0,
  .unsigned_int = 1,
  
  .table_options_str = NULL
};



static mysql_drv_args_t args;          /* driver args */

static char table_options_str[1024];   /* table options string */

static char use_ps; /* whether server-side prepared statemens should be used */

/* MySQL driver operations */

static int mysql_drv_init(void);
static int mysql_drv_describe(drv_caps_t *, const char *);
static int mysql_drv_connect(db_conn_t *);
static int mysql_drv_disconnect(db_conn_t *);
static int mysql_drv_prepare(db_stmt_t *, const char *);
static int mysql_drv_bind_param(db_stmt_t *, db_bind_t *, unsigned int);
static int mysql_drv_bind_result(db_stmt_t *, db_bind_t *, unsigned int);
static int mysql_drv_execute(db_stmt_t *, db_result_set_t *);
static int mysql_drv_fetch(db_result_set_t *);
static int mysql_drv_fetch_row(db_result_set_t *, db_row_t *);
static unsigned long long mysql_drv_num_rows(db_result_set_t *);
static int mysql_drv_query(db_conn_t *, const char *, db_result_set_t *);
static int mysql_drv_free_results(db_result_set_t *);
static int mysql_drv_close(db_stmt_t *);
static int mysql_drv_store_results(db_result_set_t *);
static int mysql_drv_done(void);

/* MySQL driver definition */

static db_driver_t mysql_driver =
{
  .sname = "mysql",
  .lname = "MySQL driver",
  .args = mysql_drv_args,
  .ops =
  {
    mysql_drv_init,
    mysql_drv_describe,
    mysql_drv_connect,
    mysql_drv_disconnect,
    mysql_drv_prepare,
    mysql_drv_bind_param,
    mysql_drv_bind_result,
    mysql_drv_execute,
    mysql_drv_fetch,
    mysql_drv_fetch_row,
    mysql_drv_num_rows,
    mysql_drv_free_results,
    mysql_drv_close,
    mysql_drv_query,
    mysql_drv_store_results,
    mysql_drv_done
  },
  .listitem = {NULL, NULL}
};


/* Local functions */

static int parse_table_type(const char *);
#ifdef HAVE_PS
static int get_mysql_bind_type(db_bind_type_t);
#endif

/* Register MySQL driver */


int register_driver_mysql(sb_list_t *drivers)
{
  SB_LIST_ADD_TAIL(&mysql_driver.listitem, drivers);

  return 0;
}


/* MySQL driver initialization */


int mysql_drv_init(void)
{
  char *s;
  
  args.host = sb_get_value_string("mysql-host");
  args.port = sb_get_value_int("mysql-port");
  args.socket = sb_get_value_string("mysql-socket");
  args.user = sb_get_value_string("mysql-user");
  args.password = sb_get_value_string("mysql-password");
  args.db = sb_get_value_string("mysql-db");
  args.myisam_max_rows = sb_get_value_int("myisam-max-rows");

  use_ps = 0;
#ifdef HAVE_PS
  mysql_drv_caps.prepared_statements = 1;
  if (db_globals.ps_mode != DB_PS_MODE_DISABLE)
    use_ps = 1;
#endif
  
  s = sb_get_value_string("mysql-table-type");

  return parse_table_type(s);
}


/* Describe database capabilities (possibly depending on table type) */


int mysql_drv_describe(drv_caps_t *caps, const char * table_name)
{
  db_conn_t    con;
  char         query[1024];
  MYSQL_RES    *res = NULL;
  MYSQL_ROW    row;
  MYSQL_FIELD  *fields;
  unsigned int num_fields;
  unsigned int i;
  int          connected = 0;
  int          rc = 0;
  
  *caps = mysql_drv_caps;
  
  if (table_name == NULL)
    goto exit;

  /* Try to determine table type */
  if (mysql_drv_connect(&con))
    goto error;
  connected = 1;
  snprintf(query, sizeof(query), "SHOW TABLE STATUS LIKE '%s'", table_name);
  if (mysql_real_query(con.ptr, query, strlen(query)))
    goto error;
  
  res = mysql_store_result(con.ptr);
  if (res == NULL)
    goto error;

  num_fields = mysql_num_fields(res);
  fields = mysql_fetch_fields(res);
  for (i = 0; i < num_fields; i++)
    if (!strcasecmp(fields[i].name, "type") ||
        !strcasecmp(fields[i].name, "engine"))
      break;
  if (i >= num_fields)
    goto error;
  
  row = mysql_fetch_row(res); 
  if (row == NULL || row[i] == NULL)
    goto error;
  
  rc = parse_table_type(row[i]);
  *caps = mysql_drv_caps;
  goto exit;

 error:
  log_text(LOG_ALERT, "Error: failed to determine table '%s' type!", table_name);
  log_text(LOG_ALERT, "MySQL error: %s", mysql_error(con.ptr));
  rc = 1;
  
 exit:
  if (res != NULL)
    mysql_free_result(res);
  if (connected)
    mysql_drv_disconnect(&con);

  return rc;
}


/* Connect to MySQL database */


int mysql_drv_connect(db_conn_t *sb_conn)
{
  MYSQL *con;

  con = (MYSQL *)malloc(sizeof(MYSQL));
  if (con == NULL)
    return 1;
  sb_conn->ptr = con;
  
  mysql_init(con);
  mysql_options(con, MYSQL_READ_DEFAULT_GROUP, "sysbench");
  if (!mysql_real_connect(con,
                         args.host,
                         args.user,
                         args.password,
                         args.db,
                         args.port,
                         args.socket,
                         0)
      )
  {
    log_text(LOG_FATAL, "unable to connect to MySQL server, aborting...");
    log_text(LOG_FATAL, "error %d: %s", mysql_errno(con),
           mysql_error(con));
    free(con);
    return 1;
  }

  return 0;
}


/* Disconnect from MySQL database */


int mysql_drv_disconnect(db_conn_t *sb_conn)
{
  MYSQL *con = sb_conn->ptr;

  if (con != NULL)
    mysql_close(con);
  free(con);
  
  return 0;
}


/* Prepare statement */


int mysql_drv_prepare(db_stmt_t *stmt, const char *query)
{
#ifdef HAVE_PS
  MYSQL      *con = (MYSQL *)stmt->connection->ptr;
  MYSQL_STMT *mystmt;

  if (con == NULL)
    return 1;

  if (use_ps)
  {
    mystmt = mysql_stmt_init(con);
    if (mystmt == NULL)
    {
      log_text(LOG_FATAL, "mysql_stmt_init() failed");
      return 1;
    }
    stmt->ptr = (void *)mystmt;
    if (mysql_stmt_prepare(mystmt, query, strlen(query)))
    {
      /* Check if this statement in not supported */
      if (mysql_errno(con) == ER_UNSUPPORTED_PS)
      {
        log_text(LOG_WARNING,
                 "Preparing of \"%s\" is unsupported, using emulation",
                 query);
        goto emulate;
      }
      else
      {
        log_text(LOG_FATAL, "mysql_stmt_prepare() failed");
        log_text(LOG_FATAL, "MySQL error: %d \"%s\"", mysql_errno(con),
                 mysql_error(con));
        mysql_stmt_close(mystmt);
        return 1;
      }
    }

    return 0;
  }

 emulate:
#endif /* HAVE_PS */

  /* Use client-side PS */
  stmt->emulated = 1;
  stmt->query = strdup(query);

  return 0;
}


/* Bind parameters for prepared statement */


int mysql_drv_bind_param(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
{
  MYSQL        *con = (MYSQL *)stmt->connection->ptr;
#ifdef HAVE_PS
  MYSQL_BIND   *bind;
  unsigned int i;
#endif
  
  if (con == NULL)
    return 1;

#ifdef HAVE_PS
  if (!stmt->emulated)
  {
    if (stmt->ptr == NULL)
      return 1;
    /* Validate parameters count */
    if (mysql_stmt_param_count(stmt->ptr) != len)
      return 1;

    /* Convert SysBench bind structures to MySQL ones */
    bind = (MYSQL_BIND *)calloc(len, sizeof(MYSQL_BIND));
    if (bind == NULL)
      return 1;
    for (i = 0; i < len; i++)
    {
      bind[i].buffer_type = get_mysql_bind_type(params[i].type);
      bind[i].buffer = params[i].buffer;
      bind[i].buffer_length = params[i].max_len;
      bind[i].length = params[i].data_len;
      bind[i].is_null = params[i].is_null;
    }

    if (mysql_stmt_bind_param(stmt->ptr, bind))
    {
      log_text(LOG_FATAL, "mysql_stmt_bind_param() failed");
      log_text(LOG_FATAL, "MySQL error: %d \"%s\"", mysql_errno(con),
                 mysql_error(con));
      free(bind);
      return 1;
    }
    free(bind);

    return 0;
  }
#endif /* HAVE_PS */

  /* Use emulation */
  if (stmt->bound_param != NULL)
    free(stmt->bound_param);
  stmt->bound_param = (db_bind_t *)malloc(len * sizeof(db_bind_t));
  if (stmt->bound_param == NULL)
    return 1;
  memcpy(stmt->bound_param, params, len * sizeof(db_bind_t));
  stmt->bound_param_len = len;
  
  return 0;

}


/* Bind results for prepared statement */


int mysql_drv_bind_result(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
{
#ifndef HAVE_PS
  /* No support for prepared statements */
  (void)stmt;   /* unused */
  (void)params; /* unused */
  (void)len;    /* unused */

  return 1;
#else
  MYSQL        *con = (MYSQL *)stmt->connection->ptr;
  MYSQL_BIND   *bind;
  unsigned int i;
  
  if (con == NULL || stmt->ptr == NULL)
    return 1;

  /* Convert SysBench bind structures to MySQL ones */
  bind = (MYSQL_BIND *)calloc(len, sizeof(MYSQL_BIND));
  if (bind == NULL)
    return 1;
  for (i = 0; i < len; i++)
  {
    bind[i].buffer_type = get_mysql_bind_type(params[i].type);
    bind[i].buffer = params[i].buffer;
    bind[i].buffer_length = params[i].max_len;
    bind[i].length = params[i].data_len;
    bind[i].is_null = params[i].is_null;
  }

  if (mysql_stmt_bind_result(stmt->ptr, bind))
  {
    free(bind);
    return 1;
  }
  free(bind);

  return 0;
#endif /* HAVE_PS */
}


/* Execute prepared statement */


int mysql_drv_execute(db_stmt_t *stmt, db_result_set_t *rs)
{
  db_conn_t       *con = stmt->connection;
  char            *buf = NULL;
  unsigned int    buflen = 0;
  unsigned int    i, j, vcnt;
  char            need_realloc;
  int             n;
  unsigned int    rc;
 
#ifdef HAVE_PS
  (void)rs; /* unused */

  if (!stmt->emulated)
  {
    if (stmt->ptr == NULL)
    {
      log_text(LOG_DEBUG, "ERROR: exiting mysql_drv_execute(), uninitialized statement");
      return SB_DB_ERROR_FAILED;
    }
    if (mysql_stmt_execute(stmt->ptr))
    {
      rc = mysql_errno(con->ptr);
      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT)
        return SB_DB_ERROR_DEADLOCK;
      log_text(LOG_ALERT, "failed to execute mysql_stmt_execute(): Err%d %s",
               mysql_errno(con->ptr),
               mysql_error(con->ptr));
      return SB_DB_ERROR_FAILED;
    }
    return SB_DB_ERROR_NONE;
  }
#else
  (void)rc; /* unused */
#endif /* HAVE_PS */

  /* Use emulation */
  /* Build the actual query string from parameters list */
  need_realloc = 1;
  vcnt = 0;
  for (i = 0, j = 0; stmt->query[i] != '\0'; i++)
  {
  again:
    if (j+1 >= buflen || need_realloc)
    {
      buflen = (buflen > 0) ? buflen * 2 : 256;
      buf = realloc(buf, buflen);
      if (buf == NULL)
      {
        log_text(LOG_DEBUG, "ERROR: exiting mysql_drv_execute(), memory allocation failure");
        return SB_DB_ERROR_FAILED;
      }
      need_realloc = 0;
    }

    if (stmt->query[i] != '?')
    {
      buf[j++] = stmt->query[i];
      continue;
    }

    n = db_print_value(stmt->bound_param + vcnt, buf + j, buflen - j);
    if (n < 0)
    {
      need_realloc = 1;
      goto again;
    }
    j += n;
    vcnt++;
  }
  buf[j] = '\0';
  
  con->db_errno = mysql_drv_query(con, buf, rs);
  free(buf);
  if (con->db_errno != SB_DB_ERROR_NONE)
  {
    log_text(LOG_DEBUG, "ERROR: exiting mysql_drv_execute(), database error");
    return con->db_errno;
  }
  
  return SB_DB_ERROR_NONE;
}


/* Execute SQL query */


int mysql_drv_query(db_conn_t *sb_conn, const char *query,
                      db_result_set_t *rs)
{
  MYSQL *con = sb_conn->ptr;
  unsigned int rc;

  (void)rs; /* unused */
  if (mysql_real_query(con, query, strlen(query)))
  {
    rc = mysql_errno(con);
    if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT)
      return SB_DB_ERROR_DEADLOCK;
    log_text(LOG_ALERT, "failed to execute MySQL query: `%s`:", query);
    log_text(LOG_ALERT, "Error %d %s", mysql_errno(con), mysql_error(con));
    return SB_DB_ERROR_FAILED; 
  }  

  return SB_DB_ERROR_NONE;
}


/* Fetch row from result set of a prepared statement */


int mysql_drv_fetch(db_result_set_t *rs)
{
  /* NYI */
  (void)rs;

  return 1;
}


/* Fetch row from result set of a query */


int mysql_drv_fetch_row(db_result_set_t *rs, db_row_t *row)
{
  /* NYI */
  (void)rs;  /* unused */
  (void)row; /* unused */
  
  return 1;
}


/* Return the number of rows in a result set */


unsigned long long mysql_drv_num_rows(db_result_set_t *rs)
{
  return rs->nrows;
}


/* Store results from the last query */


int mysql_drv_store_results(db_result_set_t *rs)
{
  MYSQL     *con = rs->connection->ptr;
  MYSQL_RES *res;

#ifdef HAVE_PS
  /* Is this result set from prepared statement? */
  if (rs->statement != NULL && rs->statement->emulated == 0)
  {
    if (rs->statement->ptr == NULL)
      return 1;
    else if (mysql_stmt_store_result(rs->statement->ptr))
    {
      log_text(LOG_ALERT, "MySQL error: %s\n", mysql_error(con));
      return 1;
    }
    rs->nrows = mysql_stmt_num_rows(rs->statement->ptr);
    while(mysql_stmt_fetch(rs->statement->ptr));

    return 0;
  }
#endif
  
  if (con == NULL)
    return 1;
  
  /* using store results for speed will not work for large sets */
  res = mysql_store_result(con);
  if (res == NULL)
  {
    log_text(LOG_ALERT, "MySQL error: %s", mysql_error(con));
    return 1; 
  }
  rs->ptr = (void *)res;

  rs->nrows = mysql_num_rows(res);
  
  /* just fetch result */
  while(mysql_fetch_row(res)); 

  return 0;
}


/* Free result set */


int mysql_drv_free_results(db_result_set_t *rs)
{
#ifdef HAVE_PS
  /* Is this a result set of a prepared statement */
  if (rs->statement != NULL && rs->statement->emulated == 0)
    return mysql_stmt_free_result(rs->statement->ptr);
#endif

  if (rs->ptr != NULL)
  {
    mysql_free_result((MYSQL_RES *)rs->ptr);
    return 0;
  }

  return 1;
}


/* Close prepared statement */


int mysql_drv_close(db_stmt_t *stmt)
{
#ifdef HAVE_PS
  if (stmt->ptr == NULL)
    return 1;
  
  return mysql_stmt_close(stmt->ptr);
#else
  (void)stmt; /* unused */
  return 0;
#endif
}


/* Uninitialize driver */
int mysql_drv_done(void)
{
  return 0;
}


/* Parse table type and set driver capabilities */


int parse_table_type(const char *type)
{
  if (!strcasecmp(type, "myisam"))
  {
    args.table_type = TABLE_TYPE_MYISAM;
    mysql_drv_caps.transactions = 0;
    if (args.myisam_max_rows > 0)
      snprintf(table_options_str, sizeof(table_options_str),
               "/*! TYPE=MyISAM MAX_ROWS=%u */", args.myisam_max_rows);
    else
      snprintf(table_options_str, sizeof(table_options_str),
             "/*! TYPE=MyISAM */");
  }
  else if (!strcasecmp(type, "innodb"))
  {
    args.table_type = TABLE_TYPE_INNODB;
    mysql_drv_caps.transactions = 1;
    snprintf(table_options_str, sizeof(table_options_str),
             "/*! TYPE=InnoDB */");
  }
  else if (!strcasecmp(type, "bdb"))
  {
    args.table_type = TABLE_TYPE_BDB;
    mysql_drv_caps.transactions = 1;
    snprintf(table_options_str, sizeof(table_options_str),
             "/*! TYPE=BDB */");
  }
  else if (!strcasecmp(type, "heap"))
  {
    args.table_type = TABLE_TYPE_HEAP;
    mysql_drv_caps.transactions = 0;
    snprintf(table_options_str, sizeof(table_options_str),
             "/*! TYPE=HEAP */");
  }
  else if (!strcasecmp(type, "ndbcluster"))
  {
    args.table_type = TABLE_TYPE_NDB;
    mysql_drv_caps.transactions = 1;
    snprintf(table_options_str, sizeof(table_options_str),
             "/*! TYPE=NDB */");
  }
  else if (!strcasecmp(type, "federated"))
  {
    args.table_type = TABLE_TYPE_FEDERATED;
    mysql_drv_caps.transactions = 1;
    snprintf(table_options_str, sizeof(table_options_str),
             "/*! TYPE=FEDERATED */");
  }
  else
  {
    log_text(LOG_FATAL, "Invalid MySQL table type: %s.", type);
    return 1;
  }

  mysql_drv_caps.table_options_str = table_options_str;
  
  return 0;
}


#ifdef HAVE_PS
/* Map SQL data type to bind_type value in MYSQL_BIND */


int get_mysql_bind_type(db_bind_type_t type)
{
  unsigned int i;

  for (i = 0; db_mysql_bind_map[i].db_type != DB_TYPE_NONE; i++)
    if (db_mysql_bind_map[i].db_type == type)
      return db_mysql_bind_map[i].my_type;

  return -1;
}

#endif /* HAVE_PS */
